# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import time
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import sonnet as snt
import tensorflow as tf
import tree
import trfl
from acme.tf import utils as tf2_utils
from acme.types import NestedArray
from acme.utils import counting, loggers

import mava
from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.components.tf.modules.exploration.exploration_scheduling import (
    LinearExplorationScheduler,
)
from mava.systems.tf import savers as tf2_savers
from mava.utils import training_utils as train_utils

train_utils.set_growing_gpu_memory()


class MADQNTrainer(mava.Trainer):
    """MADQN trainer.
    This is the trainer component of a MADQN system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        q_networks: Dict[str, snt.Module],
        target_q_networks: Dict[str, snt.Module],
        target_update_period: int,
        dataset: tf.data.Dataset,
        optimizer: Union[Dict[str, snt.Optimizer], snt.Optimizer],
        discount: float,
        shared_weights: bool,
        exploration_scheduler: LinearExplorationScheduler,
        max_gradient_norm: float = None,
        fingerprint: bool = False,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
        communication_module: Optional[BaseCommunicationModule] = None,
    ):

        self._agents = agents
        self._agent_types = agent_types
        self._shared_weights = shared_weights
        self._checkpoint = checkpoint

        # Store online and target q-networks.
        self._q_networks = q_networks
        self._target_q_networks = target_q_networks

        # General learner book-keeping and loggers.
        self._counter = counter or counting.Counter()
        self._logger = logger

        # Other learner parameters.
        self._discount = discount
        # Set up gradient clipping.
        if max_gradient_norm is not None:
            self._max_gradient_norm = tf.convert_to_tensor(max_gradient_norm)
        else:  # A very large number. Infinity results in NaNs.
            self._max_gradient_norm = tf.convert_to_tensor(1e10)

        self._fingerprint = fingerprint

        # Necessary to track when to update target networks.
        self._num_steps = tf.Variable(0, dtype=tf.int32)
        self._target_update_period = target_update_period

        # Create an iterator to go through the dataset.
        self._iterator = dataset

        # Store the exploration scheduler
        self._exploration_scheduler = exploration_scheduler

        # Dictionary with network keys for each agent.
        self.agent_net_keys = {agent: agent for agent in self._agents}
        if self._shared_weights:
            self.agent_net_keys = {agent: agent.split("_")[0] for agent in self._agents}

        self.unique_net_keys = self._agent_types if shared_weights else self._agents

        # Create optimizers for different agent types.
        if not isinstance(optimizer, dict):
            self._optimizers: Dict[str, snt.Optimizer] = {}
            for agent in self.unique_net_keys:
                self._optimizers[agent] = copy.deepcopy(optimizer)
        else:
            self._optimizers = optimizer

        # Expose the variables.
        q_networks_to_expose = {}
        self._system_network_variables: Dict[str, Dict[str, snt.Module]] = {
            "q_network": {},
        }
        for agent_key in self.unique_net_keys:
            q_network_to_expose = self._target_q_networks[agent_key]

            q_networks_to_expose[agent_key] = q_network_to_expose

            self._system_network_variables["q_network"][
                agent_key
            ] = q_network_to_expose.variables

        # Checkpointer
        self._system_checkpointer = {}
        if checkpoint:
            for agent_key in self.unique_net_keys:

                checkpointer = tf2_savers.Checkpointer(
                    directory=checkpoint_subpath,
                    time_delta_minutes=15,
                    objects_to_save={
                        "counter": self._counter,
                        "q_network": self._q_networks[agent_key],
                        "target_q_network": self._target_q_networks[agent_key],
                        "optimizer": self._optimizers,
                        "num_steps": self._num_steps,
                    },
                    enable_checkpointing=checkpoint,
                )

                self._system_checkpointer[agent_key] = checkpointer

        # Do not record timestamps until after the first learning step is done.
        # This is to avoid including the time it takes for actors to come online and
        # fill the replay buffer.

        self._timestamp: Optional[float] = None

    def get_epsilon(self) -> float:
        return self._exploration_scheduler.get_epsilon()

    def get_trainer_steps(self) -> float:
        return self._num_steps.numpy()

    def _decrement_epsilon(self) -> None:
        self._exploration_scheduler.decrement_epsilon()

    def _update_target_networks(self) -> None:
        for key in self.unique_net_keys:
            # Update target network.
            online_variables = (*self._q_networks[key].variables,)

            target_variables = (*self._target_q_networks[key].variables,)

            # Make online -> target network update ops.
            if tf.math.mod(self._num_steps, self._target_update_period) == 0:
                for src, dest in zip(online_variables, target_variables):
                    dest.assign(src)
        self._num_steps.assign_add(1)

    def _get_feed(
        self,
        o_tm1_trans: Dict[str, np.ndarray],
        o_t_trans: Dict[str, np.ndarray],
        a_tm1: Dict[str, np.ndarray],
        agent: str,
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:

        # Decentralised
        o_tm1_feed = o_tm1_trans[agent].observation
        o_t_feed = o_t_trans[agent].observation
        a_tm1_feed = a_tm1[agent]

        return o_tm1_feed, o_t_feed, a_tm1_feed

    def step(self) -> None:
        # Run the learning step.
        fetches = self._step()

        # Compute elapsed time.
        timestamp = time.time()
        if self._timestamp:
            elapsed_time = timestamp - self._timestamp
        else:
            elapsed_time = 0
        self._timestamp = timestamp  # type: ignore

        # Update our counts and record it.
        counts = self._counter.increment(steps=1, walltime=elapsed_time)
        fetches.update(counts)

        # Checkpoint and attempt to write the logs.
        if self._checkpoint:
            train_utils.checkpoint_networks(self._system_checkpointer)

        # Log and decrement epsilon
        epsilon = self.get_epsilon()
        fetches["epsilon"] = epsilon
        self._decrement_epsilon()

        if self._logger:
            self._logger.write(fetches)

    @tf.function
    def _step(
        self,
    ) -> Dict[str, Dict[str, Any]]:

        # Update the target networks
        self._update_target_networks()

        # Get data from replay (dropping extras if any). Note there is no
        # extra data here because we do not insert any into Reverb.
        inputs = next(self._iterator)

        self._forward(inputs)

        self._backward()

        # Log losses per agent
        return self._q_network_losses

    def _forward(self, inputs: Any) -> None:
        # Unpack input data as follows:
        # o_tm1 = dictionary of observations one for each agent
        # a_tm1 = dictionary of actions taken from obs in o_tm1
        # r_t = dictionary of rewards or rewards sequences
        #   (if using N step transitions) ensuing from actions a_tm1
        # d_t = environment discount ensuing from actions a_tm1.
        #   This discount is applied to future rewards after r_t.
        # o_t = dictionary of next observations or next observation sequences
        # e_t [Optional] = extra data that the agents persist in replay.
        o_tm1, a_tm1, e_tm1, r_t, d_t, o_t, e_t = inputs.data
        with tf.GradientTape(persistent=True) as tape:
            q_network_losses: Dict[str, NestedArray] = {}

            for agent in self._agents:
                agent_key = self.agent_net_keys[agent]

                # Cast the additional discount to match the environment discount dtype.
                discount = tf.cast(self._discount, dtype=d_t[agent].dtype)

                # Maybe transform the observation before feeding into policy and critic.
                # Transforming the observations this way at the start of the learning
                # step effectively means that the policy and critic share observation
                # network weights.

                o_tm1_feed, o_t_feed, a_tm1_feed = self._get_feed(
                    o_tm1, o_t, a_tm1, agent
                )

                if self._fingerprint:
                    f_tm1 = e_tm1["fingerprint"]
                    f_tm1 = tf.convert_to_tensor(f_tm1)
                    f_tm1 = tf.cast(f_tm1, "float32")

                    f_t = e_t["fingerprint"]
                    f_t = tf.convert_to_tensor(f_t)
                    f_t = tf.cast(f_t, "float32")

                    q_tm1 = self._q_networks[agent_key](o_tm1_feed, f_tm1)
                    q_t_value = self._target_q_networks[agent_key](o_t_feed, f_t)
                    q_t_selector = self._q_networks[agent_key](o_t_feed, f_t)
                else:
                    q_tm1 = self._q_networks[agent_key](o_tm1_feed)
                    q_t_value = self._target_q_networks[agent_key](o_t_feed)
                    q_t_selector = self._q_networks[agent_key](o_t_feed)

                # Q-network learning
                loss, _ = trfl.double_qlearning(
                    q_tm1,
                    a_tm1_feed,
                    r_t[agent],
                    discount * d_t[agent],
                    q_t_value,
                    q_t_selector,
                )

                loss = tf.reduce_mean(loss)

                q_network_losses[agent] = {"q_value_loss": loss}
        self._q_network_losses = q_network_losses
        self.tape = tape

    def _backward(self) -> None:
        q_network_losses = self._q_network_losses
        tape = self.tape
        for agent in self._agents:
            agent_key = self.agent_net_keys[agent]

            # Get trainable variables
            q_network_variables = self._q_networks[agent_key].trainable_variables

            # Compute gradients
            gradients = tape.gradient(q_network_losses[agent], q_network_variables)

            # Clip gradients.
            gradients = tf.clip_by_global_norm(gradients, self._max_gradient_norm)[0]

            # Apply gradients.
            self._optimizers[agent_key].apply(gradients, q_network_variables)

        train_utils.safe_del(self, "tape")

    def get_variables(self, names: Sequence[str]) -> Dict[str, Dict[str, np.ndarray]]:
        variables: Dict[str, Dict[str, np.ndarray]] = {}
        for network_type in names:
            variables[network_type] = {
                agent: tf2_utils.to_numpy(
                    self._system_network_variables[network_type][agent]
                )
                for agent in self.unique_net_keys
            }
        return variables


class MADQNRecurrentTrainer(MADQNTrainer):
    """Recurrent MADQN trainer.
    This is the trainer component of a MADQN system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        q_networks: Dict[str, snt.Module],
        target_q_networks: Dict[str, snt.Module],
        target_update_period: int,
        dataset: tf.data.Dataset,
        optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        discount: float,
        shared_weights: bool,
        exploration_scheduler: LinearExplorationScheduler,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        fingerprint: bool = False,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
        communication_module: Optional[BaseCommunicationModule] = None,
    ):
        super().__init__(
            agents=agents,
            agent_types=agent_types,
            q_networks=q_networks,
            target_q_networks=target_q_networks,
            target_update_period=target_update_period,
            dataset=dataset,
            optimizer=optimizer,
            discount=discount,
            shared_weights=shared_weights,
            exploration_scheduler=exploration_scheduler,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            logger=logger,
            fingerprint=fingerprint,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )

    def _forward(self, inputs: Any) -> None:
        data = tree.map_structure(
            lambda v: tf.expand_dims(v, axis=0) if len(v.shape) <= 1 else v, inputs.data
        )
        data = tf2_utils.batch_to_sequence(data)

        observations, actions, rewards, discounts, _, extra = data

        # core_states = extra["core_states"]
        core_state = tree.map_structure(
            lambda s: s[:, 0, :], inputs.data.extras["core_states"]
        )

        with tf.GradientTape(persistent=True) as tape:
            q_network_losses: Dict[str, NestedArray] = {}

            for agent in self._agents:
                agent_key = self.agent_net_keys[agent]
                # Cast the additional discount to match the environment discount dtype.
                discount = tf.cast(self._discount, dtype=discounts[agent][0].dtype)

                q, s = snt.static_unroll(
                    self._q_networks[agent_key],
                    observations[agent].observation,
                    core_state[agent][0],
                )

                q_targ, s = snt.static_unroll(
                    self._target_q_networks[agent_key],
                    observations[agent].observation,
                    core_state[agent][0],
                )

                q_network_losses[agent] = {"q_value_loss": tf.zeros(())}
                for t in range(1, q.shape[0]):
                    loss, _ = trfl.qlearning(
                        q[t - 1],
                        actions[agent][t - 1],
                        rewards[agent][t],
                        discount * discounts[agent][t],
                        q_targ[t],
                    )

                    loss = tf.reduce_mean(loss)
                    q_network_losses[agent]["q_value_loss"] += loss

        self._q_network_losses = q_network_losses
        self.tape = tape


class MADQNRecurrentCommTrainer(MADQNTrainer):
    """Recurrent MADQN trainer.
    This is the trainer component of a MADQN system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        q_networks: Dict[str, snt.Module],
        target_q_networks: Dict[str, snt.Module],
        target_update_period: int,
        dataset: tf.data.Dataset,
        optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        discount: float,
        shared_weights: bool,
        exploration_scheduler: LinearExplorationScheduler,
        communication_module: BaseCommunicationModule,
        max_gradient_norm: float = None,
        fingerprint: bool = False,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
    ):
        super().__init__(
            agents=agents,
            agent_types=agent_types,
            q_networks=q_networks,
            target_q_networks=target_q_networks,
            target_update_period=target_update_period,
            dataset=dataset,
            optimizer=optimizer,
            discount=discount,
            shared_weights=shared_weights,
            exploration_scheduler=exploration_scheduler,
            max_gradient_norm=max_gradient_norm,
            fingerprint=fingerprint,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )

        self._communication_module = communication_module

    def _forward(self, inputs: Any) -> None:
        data = tree.map_structure(
            lambda v: tf.expand_dims(v, axis=0) if len(v.shape) <= 1 else v, inputs.data
        )
        data = tf2_utils.batch_to_sequence(data)

        observations, actions, rewards, discounts, _, extra = data

        # core_states = extra["core_states"]
        core_state = tree.map_structure(
            lambda s: s[:, 0, :], inputs.data.extras["core_states"]
        )
        core_message = tree.map_structure(
            lambda s: s[:, 0, :], inputs.data.extras["core_messages"]
        )

        with tf.GradientTape(persistent=True) as tape:
            q_network_losses: Dict[str, NestedArray] = {
                agent: {"q_value_loss": tf.zeros(())} for agent in self._agents
            }

            T = actions[self._agents[0]].shape[0]

            state = {agent: core_state[agent][0] for agent in self._agents}
            target_state = {agent: core_state[agent][0] for agent in self._agents}

            message = {agent: core_message[agent][0] for agent in self._agents}
            target_message = {agent: core_message[agent][0] for agent in self._agents}

            # _target_q_networks must be 1 step ahead
            target_channel = self._communication_module.process_messages(target_message)
            for agent in self._agents:
                agent_key = self.agent_net_keys[agent]
                (q_targ, m), s = self._target_q_networks[agent_key](
                    observations[agent].observation[0],
                    target_state[agent],
                    target_channel[agent],
                )
                target_state[agent] = s
                target_message[agent] = m

            for t in range(1, T, 1):
                channel = self._communication_module.process_messages(message)
                target_channel = self._communication_module.process_messages(
                    target_message
                )

                for agent in self._agents:
                    agent_key = self.agent_net_keys[agent]

                    # Cast the additional discount
                    # to match the environment discount dtype.

                    discount = tf.cast(self._discount, dtype=discounts[agent][0].dtype)

                    (q_targ, m), s = self._target_q_networks[agent_key](
                        observations[agent].observation[t],
                        target_state[agent],
                        target_channel[agent],
                    )
                    target_state[agent] = s
                    target_message[agent] = m

                    (q, m), s = self._q_networks[agent_key](
                        observations[agent].observation[t - 1],
                        state[agent],
                        channel[agent],
                    )
                    state[agent] = s
                    message[agent] = m

                    loss, _ = trfl.qlearning(
                        q,
                        actions[agent][t - 1],
                        rewards[agent][t - 1],
                        discount * discounts[agent][t],
                        q_targ,
                    )

                    loss = tf.reduce_mean(loss)
                    q_network_losses[agent]["q_value_loss"] += loss

        self._q_network_losses = q_network_losses
        self.tape = tape



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps a Flatland MARL environment to be used as a dm_env environment."""


import types as tp
from functools import partial
from typing import Any, Callable, Dict, List, Sequence, Tuple, Union

import dm_env
import numpy as np
from acme import specs
from acme.wrappers.gym_wrapper import _convert_to_spec
from flatland.envs.observations import GlobalObsForRailEnv, Node, TreeObsForRailEnv
from flatland.envs.rail_env import RailEnv
from gym.spaces import Discrete
from gym.spaces.box import Box

from mava.types import OLT, Observation
from mava.utils.wrapper_utils import convert_np_type, parameterized_restart
from mava.wrappers.env_wrappers import ParallelEnvWrapper


class FlatlandEnvWrapper(ParallelEnvWrapper):
    """Environment wrapper for Flatland environments.

    All environments would require an observation preprocessor, except for
    'GlobalObsForRailEnv'. This is because flatland gives users the
    flexibility of designing custom observation builders. 'TreeObsForRailEnv'
    would use the normalize_observation function from the flatland baselines
    if none is supplied.

    The supplied preprocessor should return either an array, tuple of arrays or
    a dictionary of arrays for an observation input.

    The obervation, for an agent, returned by this wrapper could consist of both
    the agent observation and agent info. This is because flatland also provides
    informationn about the agents at each step. This information include;
    'action_required', 'malfunction', 'speed', and 'status', and it can be appended
    to the observation, by this wrapper, as an array. action_required is a boolean,
    malfunction is an int denoting the number of steps for which the agent would
    remain motionless, speed is a float and status can be any of the below;

    READY_TO_DEPART = 0
    ACTIVE = 1
    DONE = 2
    DONE_REMOVED = 3

    This would be included in the observation if agent_info is set to True
    """

    # Note: we don't inherit from base.EnvironmentWrapper because that class
    # assumes that the wrapped environment is a dm_env.Environment.
    def __init__(
        self,
        environment: RailEnv,
        preprocessor: Callable[
            [Any], Union[np.ndarray, Tuple[np.ndarray], Dict[str, np.ndarray]]
        ] = None,
        agent_info: bool = True,
    ):
        self._environment = environment
        decorate_step_method(self._environment)

        self._agents = [get_agent_id(i) for i in range(self.num_agents)]
        self._possible_agents = self.agents[:]

        self._reset_next_step = True
        self._step_type = dm_env.StepType.FIRST
        self.num_actions = 5

        self.action_spaces = {
            agent: Discrete(self.num_actions) for agent in self.possible_agents
        }

        # preprocessor must be for observation builders other than global obs
        # treeobs builders would use the default preprocessor if none is
        # supplied
        self.preprocessor: Callable[
            [Dict[int, Any]], Dict[int, Any]
        ] = self._obtain_preprocessor(preprocessor)

        self._include_agent_info = agent_info

        # observation space:
        # flatland defines no observation space for an agent. Here we try
        # to define the observation space. All agents are identical and would
        # have the same observation space.
        # Infer observation space based on returned observation
        obs, _ = self._environment.reset()
        obs = self.preprocessor(obs)
        self.observation_spaces = {
            get_agent_id(i): infer_observation_space(ob) for i, ob in obs.items()
        }

    @property
    def agents(self) -> List[str]:
        return self._agents

    @property
    def possible_agents(self) -> List[str]:
        return self._possible_agents

    def env_done(self) -> bool:
        return self._environment.dones["__all__"] or not self.agents

    def reset(self) -> dm_env.TimeStep:
        """Resets the episode."""
        self._reset_next_step = False
        self._agents = self.possible_agents[:]
        self._discounts = {
            agent: np.dtype("float32").type(1.0) for agent in self.agents
        }
        observe, info = self._environment.reset()
        observations = self._create_observations(observe, info, self._environment.dones)
        rewards_spec = self.reward_spec()
        rewards = {
            agent: convert_np_type(rewards_spec[agent].dtype, 0)
            for agent in self.possible_agents
        }

        discount_spec = self.discount_spec()
        self._discounts = {
            agent: convert_np_type(discount_spec[agent].dtype, 1)
            for agent in self.possible_agents
        }
        return parameterized_restart(rewards, self._discounts, observations)

    def step(self, actions: Dict[str, np.ndarray]) -> dm_env.TimeStep:
        """Steps the environment."""
        self._pre_step()

        if self._reset_next_step:
            return self.reset()

        self._agents = [
            agent
            for agent in self.agents
            if not self._environment.dones[get_agent_handle(agent)]
        ]

        observations, rewards, dones, infos = self._environment.step(actions)

        rewards_spec = self.reward_spec()
        #  Handle empty rewards
        if not rewards:
            rewards = {
                agent: convert_np_type(rewards_spec[agent].dtype, 0)
                for agent in self.possible_agents
            }
        else:
            rewards = {
                get_agent_id(agent): convert_np_type(
                    rewards_spec[get_agent_id(agent)].dtype, reward
                )
                for agent, reward in rewards.items()
            }

        if observations:
            observations = self._create_observations(observations, infos, dones)

        if self.env_done():
            self._step_type = dm_env.StepType.LAST
            self._reset_next_step = True
        else:
            self._step_type = dm_env.StepType.MID

        return dm_env.TimeStep(
            observation=observations,
            reward=rewards,
            discount=self._discounts,
            step_type=self._step_type,
        )

    # Convert Flatland observation so it's dm_env compatible. Also, the list
    # of legal actions must be converted to a legal actions mask.
    def _convert_observations(
        self, observes: Dict[str, Tuple[np.array, np.ndarray]], dones: Dict[str, bool]
    ) -> Observation:
        observations: Dict[str, OLT] = {}
        for agent, observation in observes.items():
            if isinstance(observation, dict) and "action_mask" in observation:
                legals = observation["action_mask"]
                observation = observation["observation"]
            else:
                # TODO Handle legal actions better for continous envs,
                #  maybe have min and max for each action and clip the agents actions
                #  accordingly
                legals = np.ones(
                    self.action_spaces[agent].shape,
                    dtype=self.action_spaces[agent].dtype,
                )
            observations[agent] = OLT(
                observation=observation,
                legal_actions=legals,
                terminal=np.asarray([dones[agent]], dtype=np.float32),
            )

        return observations

    # collate agent info and observation into a tuple, making the agents obervation to
    # be a tuple of the observation from the env and the agent info
    def _collate_obs_and_info(
        self, observes: Dict[int, np.ndarray], info: Dict[str, Dict[int, Any]]
    ) -> Dict[str, Tuple[np.array, np.ndarray]]:
        observations: Dict[str, Tuple[np.array, np.ndarray]] = {}
        observes = self.preprocessor(observes)
        for agent, obs in observes.items():
            agent_id = get_agent_id(agent)
            agent_info = np.array(
                [info[k][agent] for k in info.keys()], dtype=np.float32
            )
            obs = (obs, agent_info) if self._include_agent_info else obs
            observations[agent_id] = obs

        return observations

    def _create_observations(
        self,
        obs: Dict[int, np.ndarray],
        info: Dict[str, Dict[int, Any]],
        dones: Dict[int, bool],
    ) -> Observation:
        observations_ = self._collate_obs_and_info(obs, info)
        dones_ = {get_agent_id(k): v for k, v in dones.items()}
        observations = self._convert_observations(observations_, dones_)
        return observations

    def _obtain_preprocessor(
        self, preprocessor: Any
    ) -> Callable[[Dict[int, Any]], Dict[int, np.ndarray]]:
        """Obtains the actual preprocessor to be used based on the supplied
        preprocessor and the env's obs_builder object"""
        if not isinstance(self.obs_builder, GlobalObsForRailEnv):
            _preprocessor = preprocessor if preprocessor else lambda x: x
            if isinstance(self.obs_builder, TreeObsForRailEnv):
                _preprocessor = (
                    partial(
                        normalize_observation, tree_depth=self.obs_builder.max_depth
                    )
                    if not preprocessor
                    else preprocessor
                )
            assert _preprocessor is not None
        else:

            def _preprocessor(
                x: Tuple[np.ndarray, np.ndarray, np.ndarray]
            ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
                return x

        def returned_preprocessor(obs: Dict[int, Any]) -> Dict[int, np.ndarray]:
            temp_obs = {}
            for agent_id, ob in obs.items():
                temp_obs[agent_id] = _preprocessor(ob)
            return temp_obs

        return returned_preprocessor

    # set all parameters that should be available before an environment step
    # if no available agent, then environment is done and should be reset
    def _pre_step(self) -> None:
        if not self.agents:
            self._step_type = dm_env.StepType.LAST

    def observation_spec(self) -> Dict[str, OLT]:
        observation_specs = {}
        for agent in self.agents:
            observation_specs[agent] = OLT(
                observation=tuple(
                    (
                        _convert_to_spec(self.observation_spaces[agent]),
                        agent_info_spec(),
                    )
                )
                if self._include_agent_info
                else _convert_to_spec(self.observation_spaces[agent]),
                legal_actions=_convert_to_spec(self.action_spaces[agent]),
                terminal=specs.Array((1,), np.float32),
            )
        return observation_specs

    def action_spec(self) -> Dict[str, Union[specs.DiscreteArray, specs.BoundedArray]]:
        action_specs = {}
        action_spaces = self.action_spaces
        for agent in self.possible_agents:
            action_specs[agent] = _convert_to_spec(action_spaces[agent])
        return action_specs

    def reward_spec(self) -> Dict[str, specs.Array]:
        reward_specs = {}
        for agent in self.possible_agents:
            reward_specs[agent] = specs.Array((), np.float32)
        return reward_specs

    def discount_spec(self) -> Dict[str, specs.BoundedArray]:
        discount_specs = {}
        for agent in self.possible_agents:
            discount_specs[agent] = specs.BoundedArray(
                (), np.float32, minimum=0, maximum=1.0
            )
        return discount_specs

    def extra_spec(self) -> Dict[str, specs.BoundedArray]:
        return {}

    def seed(self, seed: int = None) -> None:
        self._environment._seed(seed)

    @property
    def environment(self) -> RailEnv:
        """Returns the wrapped environment."""
        return self._environment

    @property
    def num_agents(self) -> int:
        """Returns the number of trains/agents in the flatland environment"""
        return int(self._environment.number_of_agents)

    def __getattr__(self, name: str) -> Any:
        """Expose any other attributes of the underlying environment."""
        return getattr(self._environment, name)


# Utility functions


def infer_observation_space(
    obs: Union[tuple, np.ndarray, dict]
) -> Union[Box, tuple, dict]:
    """Infer a gym Observation space from a sample observation from flatland"""
    if isinstance(obs, np.ndarray):
        return Box(
            -np.inf,
            np.inf,
            shape=obs.shape,
            dtype=obs.dtype,
        )
    elif isinstance(obs, tuple):
        return tuple(infer_observation_space(o) for o in obs)
    elif isinstance(obs, dict):
        return {key: infer_observation_space(value) for key, value in obs.items()}
    else:
        raise ValueError(
            f"Unexpected observation type: {type(obs)}. "
            f"Observation should be of either of this types "
            f"(np.ndarray, tuple, or dict)"
        )


def agent_info_spec() -> specs.BoundedArray:
    """Create the spec for the agent_info part of the observation"""
    return specs.BoundedArray((4,), dtype=np.float32, minimum=0.0, maximum=10)


def get_agent_id(handle: int) -> str:
    """Obtain the string that constitutes the agent id from an agent handle - an int"""
    return f"train_{handle}"


def get_agent_handle(id: str) -> int:
    """Obtain an agents handle given its id"""
    return int(id.split("_")[-1])


def decorate_step_method(env: RailEnv) -> None:
    """Enable the step method of the env to take action dictionaries where agent keys
    are the agent ids. Flatland uses the agent handles as keys instead. This function
    decorates the step method so that it accepts an action dict where the keys are the
    agent ids
    """
    env.step_ = env.step

    def _step(
        self: RailEnv, actions: Dict[str, Union[int, float, Any]]
    ) -> dm_env.TimeStep:
        actions_ = {get_agent_handle(k): int(v) for k, v in actions.items()}
        return self.step_(actions_)

    env.step = tp.MethodType(_step, env)


# The block of code below is obtained from the flatland starter-kit
# at https://gitlab.aicrowd.com/flatland/flatland-starter-kit/-/blob/master/
# utils/observation_utils.py
# this is done just to obtain the normalize_observation function that would
# serve as the default preprocessor for the Tree obs builder.


def max_lt(seq: Sequence, val: Any) -> Any:
    """
    Return greatest item in seq for which item < val applies.
    None is returned if seq was empty or all items in seq were >= val.
    """
    max = 0
    idx = len(seq) - 1
    while idx >= 0:
        if seq[idx] < val and seq[idx] >= 0 and seq[idx] > max:
            max = seq[idx]
        idx -= 1
    return max


def min_gt(seq: Sequence, val: Any) -> Any:
    """
    Return smallest item in seq for which item > val applies.
    None is returned if seq was empty or all items in seq were >= val.
    """
    min = np.inf
    idx = len(seq) - 1
    while idx >= 0:
        if seq[idx] >= val and seq[idx] < min:
            min = seq[idx]
        idx -= 1
    return min


def norm_obs_clip(
    obs: np.ndarray,
    clip_min: int = -1,
    clip_max: int = 1,
    fixed_radius: int = 0,
    normalize_to_range: bool = False,
) -> np.ndarray:
    """
    This function returns the difference between min and max value of an observation
    :param obs: Observation that should be normalized
    :param clip_min: min value where observation will be clipped
    :param clip_max: max value where observation will be clipped
    :return: returnes normalized and clipped observatoin
    """
    if fixed_radius > 0:
        max_obs = fixed_radius
    else:
        max_obs = max(1, max_lt(obs, 1000)) + 1

    min_obs = 0  # min(max_obs, min_gt(obs, 0))
    if normalize_to_range:
        min_obs = min_gt(obs, 0)
    if min_obs > max_obs:
        min_obs = max_obs
    if max_obs == min_obs:
        return np.clip(np.array(obs) / max_obs, clip_min, clip_max)
    norm = np.abs(max_obs - min_obs)
    return np.clip((np.array(obs) - min_obs) / norm, clip_min, clip_max)


def _split_node_into_feature_groups(
    node: Node,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    data = np.zeros(6)
    distance = np.zeros(1)
    agent_data = np.zeros(4)

    data[0] = node.dist_own_target_encountered
    data[1] = node.dist_other_target_encountered
    data[2] = node.dist_other_agent_encountered
    data[3] = node.dist_potential_conflict
    data[4] = node.dist_unusable_switch
    data[5] = node.dist_to_next_branch

    distance[0] = node.dist_min_to_target

    agent_data[0] = node.num_agents_same_direction
    agent_data[1] = node.num_agents_opposite_direction
    agent_data[2] = node.num_agents_malfunctioning
    agent_data[3] = node.speed_min_fractional

    return data, distance, agent_data


def _split_subtree_into_feature_groups(
    node: Node, current_tree_depth: int, max_tree_depth: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    if node == -np.inf:
        remaining_depth = max_tree_depth - current_tree_depth
        # reference:
        # https://stackoverflow.com/questions/515214/total-number-of-nodes-in-a-tree-data-structure
        num_remaining_nodes = int((4 ** (remaining_depth + 1) - 1) / (4 - 1))
        return (
            [-np.inf] * num_remaining_nodes * 6,
            [-np.inf] * num_remaining_nodes,
            [-np.inf] * num_remaining_nodes * 4,
        )

    data, distance, agent_data = _split_node_into_feature_groups(node)

    if not node.childs:
        return data, distance, agent_data

    for direction in TreeObsForRailEnv.tree_explored_actions_char:
        sub_data, sub_distance, sub_agent_data = _split_subtree_into_feature_groups(
            node.childs[direction], current_tree_depth + 1, max_tree_depth
        )
        data = np.concatenate((data, sub_data))
        distance = np.concatenate((distance, sub_distance))
        agent_data = np.concatenate((agent_data, sub_agent_data))

    return data, distance, agent_data


def split_tree_into_feature_groups(
    tree: Node, max_tree_depth: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    This function splits the tree into three difference arrays of values
    """
    data, distance, agent_data = _split_node_into_feature_groups(tree)

    for direction in TreeObsForRailEnv.tree_explored_actions_char:
        sub_data, sub_distance, sub_agent_data = _split_subtree_into_feature_groups(
            tree.childs[direction], 1, max_tree_depth
        )
        data = np.concatenate((data, sub_data))
        distance = np.concatenate((distance, sub_distance))
        agent_data = np.concatenate((agent_data, sub_agent_data))

    return data, distance, agent_data


def normalize_observation(
    observation: Node, tree_depth: int, observation_radius: int = 0
) -> np.ndarray:
    """
    This function normalizes the observation used by the RL algorithm
    """
    if observation is None:
        return np.zeros(
            11 * sum(np.power(4, i) for i in range(tree_depth + 1)), dtype=np.float32
        )
    data, distance, agent_data = split_tree_into_feature_groups(observation, tree_depth)

    data = norm_obs_clip(data, fixed_radius=observation_radius)
    distance = norm_obs_clip(distance, normalize_to_range=True)
    agent_data = np.clip(agent_data, -1, 1)
    normalized_obs = np.array(
        np.concatenate((np.concatenate((data, distance)), agent_data)), dtype=np.float32
    )
    return normalized_obs



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pettingzoo environment factory."""

import dm_env
import numpy as np

from mava.utils.environments.RoboCup_env.RoboCup2D_env import RoboCup2D
from mava.wrappers.robocup import RoboCupWrapper


def make_environment(
    game_name: str = "domain_randomisation",
    evaluation: bool = False,
) -> dm_env.Environment:
    """Wraps the Robocup environment with some basic preprocessing.

    Args:
        game_name: str, the name of the Robocup game setting.
        evaluation: bool, to change the behaviour during evaluation.

    Returns:
        A Robocup environment with some standard preprocessing.
    """

    # Create environment
    if game_name == "domain_randomisation":
        players_per_team = [1, 0]
    elif game_name == "reward_shaping":
        players_per_team = [1, 0]
    # elif game_name == "fixed_opponent":
    #     players_per_team = [2, 2]
    else:
        raise NotImplementedError("Game type not implemented: ", game_name)

    # TODO: Change this to better assign ports
    rand_port = np.random.randint(6000, 60000)
    robocup_env = RoboCup2D(
        game_setting=game_name,
        team_names=["Team_A", "Team_B"],
        players_per_team=players_per_team,
        render_game=False,
        include_wait=False,
        game_length=1000,
        port=rand_port,
    )
    return RoboCupWrapper(robocup_env)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# # Adapted from
# https://github.com/deepmind/acme/blob/master/acme/adders/reverb/transition.py

"""Transition adders.
This implements an N-step transition adder which collapses trajectory sequences
into a single transition, simplifying to a simple transition adder when N=1.
"""
import copy
import itertools
import operator
from typing import Any, Optional, Tuple

import numpy as np
import reverb
import tensorflow as tf
import tree
from acme import specs as acme_specs
from acme.adders.reverb import utils as acme_utils
from acme.utils import tree_utils

from mava import specs as mava_specs
from mava.adders.reverb import base
from mava.adders.reverb import utils as mava_utils


class ParallelNStepTransitionAdder(base.ReverbParallelAdder):
    """An N-step transition adder.
    This will buffer a sequence of N timesteps in order to form a single N-step
    transition which is added to reverb for future retrieval.
    For N=1 the data added to replay will be a standard one-step transition which
    takes the form:
          (s_t, a_t, r_t, d_t, s_{t+1}, e_t)
    where:
      s_t = state observation at time t
      a_t = the action taken from s_t
      r_t = reward ensuing from action a_t
      d_t = environment discount ensuing from action a_t. This discount is
          applied to future rewards after r_t.
      e_t [Optional] = extra data that the agent persists in replay.
    For N greater than 1, transitions are of the form:
          (s_t, a_t, R_{t:t+n}, D_{t:t+n}, s_{t+N}, e_t),
    where:
      s_t = State (observation) at time t.
      a_t = Action taken from state s_t.
      g = the additional discount, used by the agent to discount future returns.
      R_{t:t+n} = N-step discounted return, i.e. accumulated over N rewards:
            R_{t:t+n} := r_t + g * d_t * r_{t+1} + ...
                             + g^{n-1} * d_t * ... * d_{t+n-2} * r_{t+n-1}.
      D_{t:t+n}: N-step product of agent discounts g_i and environment
        "discounts" d_i.
            D_{t:t+n} := g^{n-1} * d_{t} * ... * d_{t+n-1},
        For most environments d_i is 1 for all steps except the last,
        i.e. it is the episode termination signal.
      s_{t+n}: The "arrival" state, i.e. the state at time t+n.
      e_t [Optional]: A nested structure of any 'extras' the user wishes to add.
    Notes:
      - At the beginning and end of episodes, shorter transitions are added.
        That is, at the beginning of the episode, it will add:
              (s_0 -> s_1), (s_0 -> s_2), ..., (s_0 -> s_n), (s_1 -> s_{n+1})
        And at the end of the episode, it will add:
              (s_{T-n+1} -> s_T), (s_{T-n+2} -> s_T), ... (s_{T-1} -> s_T).
      - We add the *first* `extra` of each transition, not the *last*, i.e.
          if extras are provided, we get e_t, not e_{t+n}.
    """

    def __init__(
        self,
        client: reverb.Client,
        n_step: int,
        discount: float,
        priority_fns: Optional[base.PriorityFnMapping] = None,
    ) -> None:
        """Creates an N-step transition adder.
        Args:
          client: A `reverb.Client` to send the data to replay through.
          n_step: The "N" in N-step transition. See the class docstring for the
            precise definition of what an N-step transition is. `n_step` must be at
            least 1, in which case we use the standard one-step transition, i.e.
            (s_t, a_t, r_t, d_t, s_t+1, e_t).
          discount: Discount factor to apply. This corresponds to the
            agent's discount in the class docstring.
          priority_fns: See docstring for BaseAdder.
        Raises:
          ValueError: If n_step is less than 1.
        """
        # Makes the additional discount a float32, which means that it will be
        # upcast if rewards/discounts are float64 and left alone otherwise.
        self._discount = tree.map_structure(np.float32, discount)

        # Creates a placeholder for the final Step, which will have zeros for every
        # member except the observation.
        self._final_step_placeholder: Optional[base.Step] = None

        super().__init__(
            client=client,
            buffer_size=n_step,
            max_sequence_length=1,
            priority_fns=priority_fns,
            use_next_extras=True,
        )

    def _write(self) -> None:
        # NOTE: we do not check that the buffer is of length N here. This means
        # that at the beginning of an episode we will add the initial N-1
        # transitions (of size 1, 2, ...) and at the end of an episode (when
        # called from write_last) we will write the final transitions of size (N,
        # N-1, ...). See the Note in the docstring.

        # Form the n-step transition given the steps.
        observations = self._buffer[0].observations
        actions = self._buffer[0].actions
        extras = self._buffer[0].extras
        next_observations = self._next_observations
        next_extras = self._next_extras
        self._discounts = {agent: self._discount for agent in observations.keys()}

        # Give the same tree structure to the n-step return accumulator,
        # n-step discount accumulator, and self.discount, so that they can be
        # iterated in parallel using tree.map_structure.

        # If rewards dict is empty populate with zeros
        if not self._buffer[0].rewards:
            rewards = {
                agent: np.dtype("float32").type(0.0)
                for agent in self._buffer[0].discounts.keys()
            }
        else:
            rewards = self._buffer[0].rewards
        (
            n_step_return,
            total_discount,
            self_discount,
        ) = tree_utils.broadcast_structures(
            rewards, self._buffer[0].discounts, self._discounts
        )

        # Copy total_discount, so that accumulating into it doesn't affect
        # _buffer[0].discount.
        total_discount = tree.map_structure(np.copy, total_discount)

        # Broadcast n_step_return to have the broadcasted shape of
        # reward * discount. Also copy, to avoid accumulating into
        # _buffer[0].reward.
        n_step_return = tree.map_structure(
            lambda r, d: np.copy(np.broadcast_to(r, np.broadcast(r, d).shape)),
            n_step_return,
            total_discount,
        )

        # NOTE: total discount will have one less discount than it does
        # step.discounts. This is so that when the learner/update uses an additional
        # discount we don't apply it twice. Inside the following loop we will
        # apply this right before summing up the n_step_return.
        for step in itertools.islice(self._buffer, 1, None):

            # If rewards dict is empty populate with zeros
            if not step.rewards:
                rewards = {
                    agent: np.dtype("float32").type(0.0)
                    for agent in step.discounts.keys()
                }
            else:
                rewards = step.rewards
            (
                step_discount,
                step_reward,
                total_discount,
            ) = tree_utils.broadcast_structures(step.discounts, rewards, total_discount)

            # Equivalent to: `total_discount *= self._discount`.
            tree.map_structure(operator.imul, total_discount, self_discount)

            # Equivalent to: `n_step_return += step.reward * total_discount`.
            tree.map_structure(
                lambda nsr, sr, td: operator.iadd(nsr, sr * td),
                n_step_return,
                step_reward,
                total_discount,
            )

            # Equivalent to: `total_discount *= step.discount`.
            tree.map_structure(operator.imul, total_discount, step_discount)

        if extras:
            transition: Tuple[Any, ...] = (
                observations,
                actions,
                extras,
                n_step_return,
                total_discount,
                next_observations,
                next_extras,
            )
        else:
            transition = (
                observations,
                actions,
                n_step_return,
                total_discount,
                next_observations,
            )
        # Create a list of steps.
        if self._final_step_placeholder is None:
            # utils.final_step_like is expensive (around 0.085ms) to run every time
            # so we cache its output.
            self._final_step_placeholder = mava_utils.final_step_like(
                self._buffer[0], next_observations, next_extras
            )

        if next_observations is not None:
            final_step: base.Step = self._final_step_placeholder._replace(
                observations=next_observations
            )
            steps = list(self._buffer) + [final_step]

            # Calculate the priority for this transition.
            table_priorities = acme_utils.calculate_priorities(
                self._priority_fns, steps
            )
            # Insert the transition into replay along with its priority.
            self._writer.append(transition)
            for table, priority in table_priorities.items():
                self._writer.create_item(
                    table=table, num_timesteps=1, priority=priority
                )
        else:
            raise ValueError("next_observations cannot be None here.")

    def _write_last(self) -> None:
        # Drain the buffer until there are no transitions.
        self._buffer.popleft()
        while self._buffer:
            self._write()
            self._buffer.popleft()

    @classmethod
    def signature(
        cls,
        environment_spec: mava_specs.EnvironmentSpec,
        extras_spec: tf.TypeSpec = {},
    ) -> tf.TypeSpec:

        # This function currently assumes that self._discount is a scalar.
        # If it ever becomes a nested structure and/or a np.ndarray, this method
        # will need to know its structure / shape. This is because the signature
        # discount shape is the environment's discount shape and this adder's
        # discount shape broadcasted together. Also, the reward shape is this
        # signature discount shape broadcasted together with the environment
        # reward shape. As long as self._discount is a scalar, it will not affect
        # either the signature discount shape nor the signature reward shape, so we
        # can ignore it.

        agent_specs = environment_spec.get_agent_specs()
        agents = environment_spec.get_agent_ids()
        env_extras_spec = environment_spec.get_extra_specs()
        extras_spec.update(env_extras_spec)

        obs_specs = {}
        act_specs = {}
        reward_specs = {}
        step_discount_specs = {}
        for agent in agents:

            rewards_spec, step_discounts_spec = tree_utils.broadcast_structures(
                agent_specs[agent].rewards, agent_specs[agent].discounts
            )

            rewards_spec = tree.map_structure(
                _broadcast_specs, rewards_spec, step_discounts_spec
            )
            step_discounts_spec = tree.map_structure(copy.deepcopy, step_discounts_spec)

            obs_specs[agent] = agent_specs[agent].observations
            act_specs[agent] = agent_specs[agent].actions
            reward_specs[agent] = rewards_spec
            step_discount_specs[agent] = step_discounts_spec

        transition_spec = [
            obs_specs,
            act_specs,
            extras_spec,
            reward_specs,
            step_discount_specs,
            obs_specs,  # next_observation
            extras_spec,
        ]

        return tree.map_structure_with_path(
            base.spec_like_to_tensor_spec, tuple(transition_spec)
        )


def _broadcast_specs(*args: acme_specs.Array) -> acme_specs.Array:
    """Like np.broadcast, but for specs.Array.
    Args:
      *args: one or more specs.Array instances.
    Returns:
      A specs.Array with the broadcasted shape and dtype of the specs in *args.
    """
    bc_info = np.broadcast(*tuple(a.generate_value() for a in args))
    dtype = np.result_type(*tuple(a.dtype for a in args))
    return acme_specs.Array(shape=bc_info.shape, dtype=dtype)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import dm_env
import gym
import numpy as np
from acme import types

from mava.utils.environments.RoboCup_env.RoboCup2D_env import RoboCup2D
from mava.utils.environments.RoboCup_env.robocup_utils.util_functions import (  # type: ignore # noqa: E501
    SpecWrapper,
)


class RoboCupWrapper(SpecWrapper):
    """Environment wrapper for 2D RoboCup environment."""

    # Note: we don't inherit from base.EnvironmentWrapper because that class
    # assumes that the wrapped environment is a dm_env.Environment.

    def __init__(self, environment: RoboCup2D) -> None:
        self._environment = environment
        self._reset_next_step = True
        assert environment.game_setting in ["reward_shaping", "domain_randomisation"]

        super().__init__(environment.num_players)

    def reset(self) -> Tuple[dm_env.TimeStep, np.array]:
        """Resets the episode."""
        self._reset_next_step = False
        raw_obs, _, state = self._environment.reset()
        proc_obs = self._proc_robocup_obs(observations=raw_obs, done=False)
        proccessed_state = self._proc_robocup_state(state, proc_obs)
        timestep = dm_env.restart(proc_obs)
        return timestep, {"env_state": proccessed_state}

    def step(self, nn_actions: types.NestedArray) -> Tuple[dm_env.TimeStep, np.array]:
        """Steps the environment."""
        if self._reset_next_step:
            return self.reset()

        actions = self._proc_robocup_actions(nn_actions)
        raw_obs, rewards, state, done = self._environment.step(actions)
        self._reset_next_step = done

        proc_obs = self._proc_robocup_obs(
            observations=raw_obs, done=done, nn_actions=nn_actions
        )
        proccessed_state = self._proc_robocup_state(state, proc_obs)

        if done:
            self._step_type = dm_env.StepType.LAST
        else:
            self._step_type = dm_env.StepType.MID

        return (
            dm_env.TimeStep(
                observation=proc_obs,
                reward=rewards,
                discount=self._discount,
                step_type=self._step_type,
            ),
            {"env_state": proccessed_state},
        )

    @property
    def environment(self) -> gym.Env:
        """Returns the wrapped environment."""
        return self._environment



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for logging to the terminal."""
import time
from typing import Dict, List, Optional, Type

import tensorflow as tf
from acme.utils.loggers import base
from tensorflow import Tensor


def format_key(key: str) -> str:
    """Internal function for formatting keys in Tensorboard format."""
    return key.title().replace("_", "")


def format_key_histograms(key: str) -> str:
    """Internal function for formatting keys in Tensorboard format."""
    return key.title().replace(":", "_") + "_hist"


class TFSummaryLogger(base.Logger):
    """Logs to a tf.summary created in a given logdir.
    If multiple TFSummaryLogger are created with the same logdir, results will be
    categorized by labels.
    """

    def __init__(self, logdir: str, label: str = "Logs"):
        """Initializes the logger.
        Args:
            logdir: directory to which we should log files.
            label: label string to use when logging. Default to 'Logs'.
        """
        self._time = time.time()
        self.label = label
        self._iter = 0
        self._summary: Optional[Type[tf.summary.SummaryWriter]] = None
        self._logdir = logdir

    def write(self, values: base.LoggingData) -> None:
        # If this is in init, launchpad fails,
        # Error: tensorflow.python.framework.errors_impl.InvalidArgumentError:
        #   Cannot convert a Tensor of dtype resource to a NumPy array.
        # Line: CloudPickler(file, protocol=protocol, buffer_callback
        #   =buffer_callback).dump(obj)
        if self._summary is None:
            self._summary = tf.summary.create_file_writer(self._logdir)

        with self._summary.as_default():
            for key, value in values.items():
                if hasattr(value, "shape") and len(value.shape) > 0:
                    self.histogram_summary(key, value)
                elif hasattr(value, "shape") or not isinstance(value, dict):
                    self.scalar_summary(key, value)
                else:
                    self.dict_summary(key, value)
            self._iter += 1

    def scalar_summary(self, key: str, value: float) -> None:
        tf.summary.scalar(f"{self.label}/{format_key(key)}", value, step=self._iter)

    def dict_summary(self, key: str, value: Dict) -> None:
        dict_info = self._flatten_dict(parent_key=key, dict_info=value)
        for (k, v) in dict_info.items():
            self.scalar_summary(k, v)

    def histogram_summary(self, key: str, value: Tensor) -> None:
        tf.summary.histogram(
            f"{self.label}/{format_key_histograms(key)}", value, step=self._iter
        )

    # Flatten dict, adapted from
    # https://stackoverflow.com/questions/6027558/flatten-nested-dictionaries-compressing-keys
    # Converts {'agent_0': {'critic_loss': 0.1, 'policy_loss': 0.2},...}
    #   to  {'agent_0_critic_loss':0.1,'agent_0_policy_loss':0.1 ,...}
    def _flatten_dict(
        self, parent_key: str, dict_info: Dict, sep: str = "_"
    ) -> Dict[str, float]:
        items: List = []
        for k, v in dict_info.items():
            k = str(k)
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, dict):
                items.extend(
                    self._flatten_dict(parent_key=new_key, dict_info=v, sep=sep).items()
                )
            else:
                items.append((new_key, v))
        return dict(items)



"""Utility classes for saving model checkpoints and snapshots
 that leverages acme's checkpointer."""

from typing import Any, Union

from acme import core
from acme.tf import savers


# Classes adapted from https://github.com/deepmind/acme/blob/master/acme/tf/savers.py
class Checkpointer(savers.Checkpointer):
    def __init__(self, directory: str = "~/mava/", **kwargs: Any) -> None:
        super().__init__(directory=directory, add_uid=False, **kwargs)


class CheckpointingRunner(savers.CheckpointingRunner):
    """Wrap an object and expose a run method which checkpoints periodically.

    This internally creates a Checkpointer around `wrapped` object and exposes
    all of the methods of `wrapped`. Additionally, anay `**kwargs` passed to the
    runner are forwarded to the internal Checkpointer.
    """

    def __init__(
        self,
        wrapped: Union[
            savers.Checkpointable, core.Saveable, core.Learner, savers.TFSaveable
        ],
        *,
        time_delta_minutes: int = 30,
        **kwargs: Any,
    ):

        if isinstance(wrapped, savers.TFSaveable):
            # If the object to be wrapped exposes its TF State, checkpoint that.
            objects_to_save = wrapped.state
        else:
            # Otherwise checkpoint the wrapped object itself.
            objects_to_save = wrapped

        self._wrapped = wrapped
        self._time_delta_minutes = time_delta_minutes
        self._checkpointer = Checkpointer(
            objects_to_save={"wrapped": objects_to_save},
            time_delta_minutes=time_delta_minutes,
            **kwargs,
        )


class Snapshotter(savers.Snapshotter):
    """Convenience class for periodically snapshotting."""

    def __init__(self, directory: str = "~/mava/", **kwargs: Any) -> None:
        super().__init__(directory=directory, **kwargs)



"""Distributions, for use in acme/networks/distributional.py."""
from typing import Tuple, Union

import numpy as np
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp
from acme.tf.networks.distributions import (
    DiscreteValuedDistribution as acme_DiscreteValuedDistribution,
)
from tensorflow_probability import distributions as tfd


class DiscreteValuedHead(snt.Module):
    """Represents a parameterized discrete valued distribution.

    The returned distribution is essentially a `tfd.Categorical`, but one which
    knows its support and so can compute the mean value.
    """

    def __init__(
        self,
        vmin: Union[float, np.ndarray, tf.Tensor],
        vmax: Union[float, np.ndarray, tf.Tensor],
        num_atoms: int,
        w_init: snt.initializers.Initializer = None,
        b_init: snt.initializers.Initializer = None,
    ):
        """Initialization.

        If vmin and vmax have shape S, this will store the category values as a
        Tensor of shape (S*, num_atoms).

        Args:
          vmin: Minimum of the value range
          vmax: Maximum of the value range
          num_atoms: The atom values associated with each bin.
          w_init: Initialization for linear layer weights.
          b_init: Initialization for linear layer biases.
        """
        super().__init__(name="DiscreteValuedHead")
        vmin = tf.convert_to_tensor(vmin)
        vmax = tf.convert_to_tensor(vmax)
        self._values = tf.linspace(vmin, vmax, num_atoms, axis=-1)
        self._distributional_layer = snt.Linear(
            tf.size(self._values), w_init=w_init, b_init=b_init
        )

    def __call__(self, inputs: tf.Tensor) -> tfd.Distribution:
        logits = self._distributional_layer(inputs)
        logits = tf.reshape(
            logits,
            tf.concat(
                [tf.shape(logits)[:1], tf.shape(self._values)], axis=0  # batch size
            ),
        )
        values = tf.cast(self._values, logits.dtype)

        return DiscreteValuedDistribution(values=values, logits=logits)


@tfp.experimental.register_composite
class DiscreteValuedDistribution(acme_DiscreteValuedDistribution):
    """This is a generalization of a categorical distribution.

    The support for the DiscreteValued distribution can be any real valued range,
    whereas the categorical distribution has support [0, n_categories - 1] or
    [1, n_categories]. This generalization allows us to take the mean of the
    distribution over its support.
    """

    def __init__(
        self,
        values: tf.Tensor,
        logits: tf.Tensor = None,
        probs: tf.Tensor = None,
        name: str = "DiscreteValuedDistribution",
    ):
        """Initialization.

        Args:
          values: Values making up support of the distribution. Should have a shape
            compatible with logits.
          logits: An N-D Tensor, N >= 1, representing the log probabilities of a set
            of Categorical distributions. The first N - 1 dimensions index into a
            batch of independent distributions and the last dimension indexes into
            the classes.
          probs: An N-D Tensor, N >= 1, representing the probabilities of a set of
            Categorical distributions. The first N - 1 dimensions index into a batch
            of independent distributions and the last dimension represents a vector
            of probabilities for each class. Only one of logits or probs should be
            passed in.
          name: Name of the distribution object.
        """
        self.true_dimensions = None
        super().__init__(values=values, logits=logits, probs=probs, name=name)

    def set_dimensions(self, dims: Tuple) -> tfd.Categorical:
        assert len(dims) == 2
        self._true_dimensions = [dims[0], dims[1]]

    def cut_dimension(
        self, axis: int, start: int = 0, end: Union[int, None] = None
    ) -> None:
        if end is None:
            end = self._true_dimensions[axis]
        if axis == 1:
            reshape_dims = self._true_dimensions + [self._values.shape[0]]
            self._logits: tf.Tensor = tf.reshape(
                tf.reshape(self._logits, reshape_dims)[:, start:end],
                [-1, self._values.shape[0]],
            )
            if end is not None:
                self._true_dimensions[axis] = end - start
        else:
            raise NotImplementedError



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OpenAI Gym environment factory."""

import dm_env
import gym
from acme import wrappers

TASKS = {
    "debug": ["MountainCarContinuous-v0"],
    "default": [
        "HalfCheetah-v2",
        "Hopper-v2",
        "InvertedDoublePendulum-v2",
        "InvertedPendulum-v2",
        "Reacher-v2",
        "Swimmer-v2",
        "Walker2d-v2",
    ],
}


def make_environment(
    evaluation: bool = False, task: str = "MountainCarContinuous-v0"
) -> dm_env.Environment:
    """Creates an OpenAI Gym environment."""
    del evaluation

    # Load the gym environment.
    environment = gym.make(task)

    # Make sure the environment obeys the dm_env.Environment interface.
    environment = wrappers.GymWrapper(environment)
    # Clip the action returned by the agent to the environment spec.
    environment = wrappers.CanonicalSpecWrapper(environment, clip=True)
    environment = wrappers.SinglePrecisionWrapper(environment)

    return environment



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pettingzoo environment factory."""

import importlib
from typing import Any, List, Optional, Union

import dm_env
import numpy as np
import supersuit
from supersuit import black_death_v1

from mava.wrappers import (
    ParallelEnvWrapper,
    PettingZooAECEnvWrapper,
    PettingZooParallelEnvWrapper,
    SequentialEnvWrapper,
)


def atari_preprocessing(
    env: Union[ParallelEnvWrapper, SequentialEnvWrapper]
) -> Union[ParallelEnvWrapper, SequentialEnvWrapper]:

    # Preprocessing
    env = supersuit.max_observation_v0(env, 2)

    # repeat_action_probability is set to 0.25
    # to introduce non-determinism to the system
    env = supersuit.sticky_actions_v0(env, repeat_action_probability=0.25)

    # skip frames for faster processing and less control
    # to be compatable with gym, use frame_skip(env, (2,5))
    env = supersuit.frame_skip_v0(env, 4)

    # downscale observation for faster processing
    env = supersuit.resize_v0(env, 84, 84)

    # allow agent to see everything on the screen
    # despite Atari's flickering screen problem
    env = supersuit.frame_stack_v1(env, 4)

    # set dtype to float32
    env = supersuit.dtype_v0(env, np.float32)

    return env


def make_environment(
    evaluation: bool = False,
    env_type: str = "parallel",
    env_class: str = "mpe",
    env_name: str = "simple_spread_v2",
    env_preprocess_wrappers: Optional[List] = [(black_death_v1, None)],
    random_seed: Optional[int] = None,
    **kwargs: Any,
) -> dm_env.Environment:
    """Wraps an Pettingzoo environment.

    Args:
        env_class: str, class of the environment, e.g. MPE or Atari.
        env_name: str, name of environment, .e.g simple_spread or Pong.
        evaluation: bool, to change the behaviour during evaluation.

    Returns:
        A Pettingzoo environment wrapped as a DeepMind environment.
    """
    del evaluation

    env_module = importlib.import_module(f"pettingzoo.{env_class}.{env_name}")

    if env_type == "parallel":
        env = env_module.parallel_env(**kwargs)  # type: ignore
        if env_class == "atari":
            env = atari_preprocessing(env)
        # wrap parallel environment
        environment = PettingZooParallelEnvWrapper(
            env, env_preprocess_wrappers=env_preprocess_wrappers
        )
    elif env_type == "sequential":
        env = env_module.env(**kwargs)  # type: ignore
        if env_class == "atari":
            env = atari_preprocessing(env)
        # wrap sequential environment
        environment = PettingZooAECEnvWrapper(
            env, env_preprocess_wrappers=env_preprocess_wrappers
        )

    if random_seed and hasattr(environment, "seed"):
        environment.seed(random_seed)

    return environment



# !/usr/bin/env python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import random
import threading
import time

import numpy as np

from mava.utils.environments.RoboCup_env.robocup_utils import (
    handler,
    sock,
    sp_exceptions,
)
from mava.utils.environments.RoboCup_env.robocup_utils.game_object import Flag
from mava.utils.environments.RoboCup_env.robocup_utils.trainer_world_model import (
    WorldModel,
)

max_x_rand = Flag.out_x
max_y_rand = Flag.out_y

# Define some colours
green = (0, 255, 50)
light_blue = (0, 255, 255)
blue = (0, 0, 255)
red = (255, 0, 0)
yellow = (255, 255, 0)
white = (255, 255, 255)

# R G B
field_col = (0, 200, 50)
team_player = blue
opp_player = light_blue
own_goal = (255, 140, 0)
opp_goal = (128, 0, 128)


class Trainer:
    def __init__(self):
        # whether we're connected to a server yet or not
        self.__connected = False

        # set all variables and important objects to appropriate values for
        # pre-connect state.

        # the socket used to communicate with the server
        self.__sock = None

        # models and the message handler for parsing and storing information
        self.wm = None
        self.msg_handler = None

        # parse thread and control variable
        self.__parsing = False
        self.__msg_thread = None

        self.__thinking = False  # think thread and control variable
        self.__think_thread = None

        # whether we should run the think method
        self.__should_think_on_data = False

        # whether we should send commands
        self.__send_commands = True

        # adding goal post markers
        self.enemy_goal_pos = None
        self.own_goal_pos = None

    def connect(self, host, port, version=11):
        """
        Gives us a connection to the server as one player on a team.  This
        immediately connects the agent to the server and starts receiving and
        parsing the information it sends.
        """

        # if already connected, raise an error since user may have wanted to
        # connect again to a different server.
        if self.__connected:
            msg = "Cannot connect while already connected, disconnect first."
            raise sp_exceptions.AgentConnectionStateError(msg)

        # the pipe through which all of our communication takes place
        self.__sock = sock.Socket(host, port)

        # our models of the world and our body
        self.wm = WorldModel(handler.ActionHandler(self.__sock))

        # handles all messages received from the server
        self.msg_handler = handler.MessageHandler(self.wm)

        # set up our threaded message receiving system
        self.__parsing = True  # tell thread that we're currently running
        self.__msg_thread = threading.Thread(
            target=self.__message_loop, name="message_loop"
        )
        self.__msg_thread.daemon = True  # dies when parent thread dies

        # start processing received messages. this will catch the initial server
        # response and all subsequent communication.
        self.__msg_thread.start()

        # send the init message and allow the message handler to handle further
        # responses.
        # init_address = self.__sock.address
        # teamname = "John"
        init_msg = "(init (version %d))"
        self.__sock.send(init_msg % (version))

        # # wait until the socket receives a response from the server and gets its
        # # assigned port.
        # while self.__sock.address == init_address:
        #     time.sleep(0.0001)

        # create our thinking thread.  this will perform the actions necessary
        # to play a game of robo-soccer.
        self.__thinking = False
        # self.__think_thread = threading.Thread(target=self.__think_loop,
        #         name="think_loop")
        # self.__think_thread.daemon = True

        # set connected state.  done last to prevent state inconsistency if
        # something goes wrong beforehand.
        self.__connected = True
        time.sleep(1)
        self.__sock.send("(eye on)")
        self.__sock.send("(ear on)")

    def run(self):
        """
        Kicks off the thread that does the agent's thinking, allowing it to play
        during the game.  Throws an exception if called while the agent is
        already playing.
        """

        # ensure we're connected before doing anything
        if not self.__connected:
            msg = "Must be connected to a server to begin play."
            raise sp_exceptions.AgentConnectionStateError(msg)

        # throw exception if called while thread is already running
        if self.__thinking:
            raise sp_exceptions.AgentAlreadyPlayingError("Agent is already playing.")

        # run the method that sets up the agent's persistant variables
        self.setup_environment()

        # tell the thread that it should be running, then start it
        self.__thinking = True
        self.__should_think_on_data = True
        self.__think_loop()

    # def disconnect(self):
    #     """
    #     Tell the loop threads to stop and signal the server that we're
    #     disconnecting, then join the loop threads and destroy all our inner
    #     methods.

    #     Since the message loop thread can conceiveably block indefinitely while
    #     waiting for the server to respond, we only allow it (and the think loop
    #     for good measure) a short time to finish before simply giving up.

    #     Once an agent has been disconnected, it is 'dead' and cannot be used
    #     again.  All of its methods get replaced by a method that raises an
    #     exception every time it is called.
    #     """

    #     # don't do anything if not connected
    #     if not self.__connected:
    #         return

    #     # tell the loops to terminate
    #     self.__parsing = False
    #     self.__thinking = False

    #     # tell the server that we're quitting
    #     self.__sock.send("(bye)")

    #     # tell our threads to join, but only wait breifly for them to do so.
    #     # don't join them if they haven't been started (this can happen if
    #     # disconnect is called very quickly after connect).
    #     if self.__msg_thread.is_alive():
    #         self.__msg_thread.join(0.01)

    #     # if self.__think_thread.is_alive():
    #     #     self.__think_thread.join(0.01)

    #     # reset all standard variables in this object.  self.__connected gets
    #     # reset here, along with all other non-user defined internal variables.
    #     Agent.__init__(self)

    def __message_loop(self):
        """
        Handles messages received from the server.

        This SHOULD NOT be called externally, since it's used as a threaded loop
        internally by this object.  Calling it externally is a BAD THING!
        """

        # loop until we're told to stop
        while self.__parsing:
            # receive message data from the server and pass it along to the
            # world model as-is.  the world model parses it and stores it within
            # itself for perusal at our leisure.
            raw_msg = self.__sock.recv()

            msg_type = self.msg_handler.handle_message(raw_msg)

            # if b'goal_l' in raw_msg or b'goal_r' in raw_msg:
            # print("Trainer message: '" , raw_msg,
            # ". Scores: ", self.wm.score_l, self.wm.score_r)
            # exit()

            # if msg_type is not None:
            # print(type(raw_msg))

            # we send commands all at once every cycle, ie. whenever a
            # 'sense_body' command is received
            if msg_type == handler.ActionHandler.CommandType.SENSE_BODY:
                self.__send_commands = True

            # flag new data as needing the think loop's attention
            self.__should_think_on_data = True

    # def __think_loop(self):
    #     """
    #     Performs world model analysis and sends appropriate commands to the
    #     server to allow the agent to participate in the current game.
    #
    #     Like the message loop, this SHOULD NOT be called externally.  Use the
    #     play method to start play, and the disconnect method to end it.
    #     """
    #
    #     while self.__thinking:
    #         # tell the ActionHandler to send its enqueued messages if it is time
    #         if self.__send_commands:
    #             self.__send_commands = False
    #             self.wm.ah.send_commands()
    #
    #         # only think if new data has arrived
    #         if self.__should_think_on_data:
    #             # flag that data has been processed.  this shouldn't be a race
    #             # condition, since the only change would be to make it True
    #             # before changing it to False again, and we're already going to
    #             # process data, so it doesn't make any difference.
    #             self.__should_think_on_data = False
    #
    #             self.think()
    #         else:
    #             # prevent from burning up all the cpu time while waiting for data
    #             time.sleep(0.0001)

    def setup_environment(self):
        """
        Called before the think loop starts, this allows the user to store any
        variables/objects they'll want access to across subsequent calls to the
        think method.
        """

        self.in_kick_off_formation = False

    def send_done(self):
        self.__sock.send("(done)")

    def reset_game(
        self,
        players_per_team,
        team_names,
        game_setting,
        game_diff=None,
        reset_stamina=True,
    ):
        self.__sock.send("(change_mode before_kick_off)")
        if reset_stamina:
            self.__sock.send("(recover)")
        # github.com/rcsoccersim/rcssserver/blob/master/src/coach.cpp

        # game_easy = 1 - game_diff
        if game_setting == "domain_randomisation":
            # print("game_diff: ", game_diff)
            assert game_diff is not None

            # Move the ball

            if players_per_team[0] > 0 and players_per_team[1] > 0:
                rand_side = random.randint(0, 1) * 2 - 1
            elif players_per_team[0] > 0:
                rand_side = 1
            else:
                rand_side = 0
            s_x = 52
            # TODONE: Change this back to the harder setting!
            x = str(
                np.random.randint(
                    s_x - (s_x + max_x_rand) * game_diff,
                    s_x - (s_x - max_x_rand) * game_diff + 1,
                )
                * rand_side
            )
            # str(np.random.randint(s_x - (s_x + max_x_rand)*game_diff,
            # s_x - (s_x - max_x_rand)*game_diff + 1)*rand_side)
            # str(s_x*rand_side)
            y = str(
                np.random.randint(-max_y_rand * game_diff, 1 + max_y_rand * game_diff)
            )
            # str(np.random.randint(-max_y_rand*game_diff, 1 + max_y_rand*game_diff))
            # str(0)  #
            self.__sock.send("(move (ball) " + x + " " + y + ")")
            for t_i in range(len(players_per_team)):
                for p_i in range(1, players_per_team[t_i] + 1):
                    s_x = 51
                    x = str(
                        np.random.randint(
                            s_x - (s_x + max_x_rand) * game_diff,
                            s_x - (s_x - max_x_rand) * game_diff + 1,
                        )
                        * (1 - t_i * 2)
                    )
                    y = str(
                        np.random.randint(
                            -max_y_rand * game_diff, 1 + max_y_rand * game_diff
                        )
                    )
                    ang = str(np.random.randint(-180 * game_diff, 1 + 180 * game_diff))

                    command = (
                        "(move (player "
                        + team_names[t_i]
                        + " "
                        + str(p_i)
                        + ") "
                        + str(x)
                        + " "
                        + str(y)
                        + " "
                        + str(ang)
                        + " 0 0)"
                    )
                    self.__sock.send(command)
        elif game_setting == "reward_shaping":
            # Move the ball
            x = str(np.random.randint(-max_x_rand, max_x_rand + 1))
            y = str(np.random.randint(-max_y_rand, max_y_rand + 1))
            self.__sock.send("(move (ball) " + x + " " + y + ")")

            for t, team_name in enumerate(team_names):
                omt = 1 - t
                for p_i in range(1, players_per_team[0] + 1):
                    x = str(np.random.randint(-51 * omt, 51 * t + 1))
                    y = str(np.random.randint(-max_y_rand, max_y_rand + 1))
                    ang = str(np.random.randint(-180, 180 + 1))

                    command = (
                        "(move (player "
                        + team_name
                        + " "
                        + str(p_i)
                        + ") "
                        + str(x)
                        + " "
                        + str(y)
                        + " "
                        + str(ang)
                        + " 0 0)"
                    )
                    self.__sock.send(command)

        else:
            raise NotImplementedError(
                "Game setting not implemented in trainer: ", game_setting
            )
        self.__sock.send("(change_mode play_on)")
        self.__sock.send("(start)")
        self.__sock.send("(done)")

    def get_state_dict(self):
        return self.wm.get_state()



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Core Mava interfaces.
This file specifies and documents the notions of `Executor` and `Trainer`
similar to the `Actor` and `Learner` in Acme.
"""

import abc
import itertools
from typing import Dict, Generic, Optional, Sequence, Tuple, TypeVar, Union

import acme
import dm_env
from acme import types

T = TypeVar("T")


class Executor(acme.Actor):
    """Interface for a system that can execute agent policies.
    This interface defines an API for a System to interact with an EnvironmentLoop
    (see mava.environment_loop), e.g. a simple RL loop where each step is of the
    form:
      # Make the first observation.
      timestep = env.reset()
      system.observe_first(timestep.observation)
      # Take a step and observe.
      action = system.select_actions(timestep.observation)
      next_timestep = env.step(action)
      actor.observe(action, next_timestep)
      # Update the actor policy/parameters.
      system.update()
    """

    @abc.abstractmethod
    def select_action(
        self, agent: str, observation: types.NestedArray
    ) -> types.NestedArray:
        """Samples from the policy and returns an action."""

    @abc.abstractmethod
    def select_actions(
        self, observations: Dict[str, types.NestedArray]
    ) -> Union[
        Dict[str, types.NestedArray],
        Tuple[Dict[str, types.NestedArray], Dict[str, types.NestedArray]],
    ]:
        """Samples from the policy and returns an action for each agent."""

    @abc.abstractmethod
    def observe(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Dict[str, types.NestedArray] = {},
    ) -> None:
        """Make an observation of timestep data from parallel environment.
        Args:
        action: action taken in the environment.
        next_timestep: timestep produced by the environment given the action.
        """


# Internal class.


class VariableSource(abc.ABC):
    """Abstract source of variables.
    Objects which implement this interface provide a source of variables, returned
    as a collection of (nested) numpy arrays. Generally this will be used to
    provide variables to some learned policy/etc.
    """

    @abc.abstractmethod
    def get_variables(
        self, names: Sequence[str]
    ) -> Dict[str, Dict[str, types.NestedArray]]:
        """Return the named variables as a collection of (nested) numpy arrays.
        Args:
        names: args where each name is a string identifying a predefined subset of
            the variables.
        Returns:
        A list of (nested) numpy arrays `variables` such that `variables[i]`
        corresponds to the collection named by `names[i]`.
        """


class Worker(abc.ABC):
    """An interface for (potentially) distributed workers."""

    @abc.abstractmethod
    def run(self) -> None:
        """Runs the worker."""


class Saveable(abc.ABC, Generic[T]):
    """An interface for saveable objects."""

    @abc.abstractmethod
    def save(self) -> T:
        """Returns the state from the object to be saved."""

    @abc.abstractmethod
    def restore(self, state: T) -> None:
        """Given the state, restores the object."""


class Trainer(VariableSource, Worker, Saveable):
    """Abstract learner object.
    This corresponds to an object which implements a learning loop. A single step
    of learning should be implemented via the `step` method and this step
    is generally interacted with via the `run` method which runs update
    continuously.
    All objects implementing this interface should also be able to take in an
    external dataset (see acme.datasets) and run updates using data from this
    dataset. This can be accomplished by explicitly running `learner.step()`
    inside a for/while loop or by using the `learner.run()` convenience function.
    Data will be read from this dataset asynchronously and this is primarily
    useful when the dataset is filled by an external process.
    """

    @abc.abstractmethod
    def step(self) -> None:
        """Perform an update step of the learner's parameters."""

    def run(self, num_steps: Optional[int] = None) -> None:
        """Run the update loop; typically an infinite loop which calls step."""

        iterator = range(num_steps) if num_steps is not None else itertools.count()

        for _ in iterator:
            self.step()

    def save(self) -> T:
        raise NotImplementedError('Method "save" is not implemented.')

    def restore(self, state: T) -> None:
        raise NotImplementedError('Method "restore" is not implemented.')



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Adapted from https://github.com/openai/multiagent-particle-envs.
# TODO (dries): Try using this class directly from PettingZoo and delete this file.

import copy
from typing import Any, Callable, Dict, List, Tuple, Union

import gym
import numpy as np
from gym import spaces

from mava.utils.debugging.core import Agent, World

from .multi_discrete import MultiDiscrete


# environment for all agents in the multiagent world
# currently code assumes that no agents will be created/destroyed at runtime!
class MultiAgentEnv(gym.Env):
    metadata = {"render.modes": ["human", "rgb_array"]}

    def __init__(
        self,
        world: World,
        action_space: str,
        reset_callback: Callable = None,
        reward_callback: Callable = None,
        observation_callback: Callable = None,
        info_callback: Callable = None,
        done_callback: Callable = None,
        shared_viewer: bool = True,
    ) -> None:

        self.world = world
        # Generate IDs and convert agent list to dictionary format.
        agent_dict = {}
        self.env_done = False
        self.agent_ids = []

        agent_list = self.world.policy_agents
        for a_i in range(len(agent_list)):
            agent_id = "agent_" + str(a_i)
            self.agent_ids.append(agent_id)
            agent_dict[agent_id] = agent_list[a_i]

        self.possible_agents = self.agent_ids

        self.num_agents = len(self.agent_ids)
        self.agents = agent_dict

        # set required vectorized gym env property
        self.n = len(world.policy_agents)
        # scenario callbacks
        self.reset_callback = reset_callback
        self.reward_callback = reward_callback
        self.observation_callback = observation_callback
        self.info_callback = info_callback
        self.done_callback = done_callback
        # environment parameters
        assert action_space in ["continuous", "discrete"]

        self.discrete_action_space = action_space == "discrete"

        # if true, action is a number 0...N, otherwise
        # action is a one-hot N-dimensional vector
        self.discrete_action_input = action_space == "discrete"
        self.time = 0

        self.render_geoms: Union[List, None] = None
        self.render_geoms_xform: Union[List, None] = None

        # configure spaces
        self.action_spaces = {}
        self.observation_spaces = {}
        for a_i, agent_id in enumerate(self.agent_ids):
            agent = self.agents[agent_id]
            total_action_space = []
            # physical action space
            if self.discrete_action_space:
                u_action_space = spaces.Discrete(world.dim_p * 2 + 1)
            else:
                u_action_space = spaces.Box(
                    low=-agent.u_range,
                    high=+agent.u_range,
                    shape=(world.dim_p,),
                    dtype=np.float32,
                )
            if agent.movable:
                total_action_space.append(u_action_space)
            # total action space
            if len(total_action_space) > 1:
                # all action spaces are discrete, so simplify to
                # MultiDiscrete action space
                if all(
                    isinstance(act_space, spaces.Discrete)
                    for act_space in total_action_space
                ):
                    act_space = MultiDiscrete(
                        [[0, act_space.n - 1] for act_space in total_action_space]
                    )
                else:
                    act_space = spaces.Tuple(total_action_space)
                self.action_spaces[agent_id] = act_space
            else:
                self.action_spaces[agent_id] = total_action_space[0]
            # observation space
            if observation_callback is not None:
                obs_dim = len(observation_callback(agent, a_i, self.world))
            else:
                raise ValueError("Observation callback is None.")
            self.observation_spaces[agent_id] = spaces.Box(
                low=-np.inf, high=+np.inf, shape=(obs_dim,), dtype=np.float32
            )

        # rendering
        self.shared_viewer = shared_viewer
        self.viewers: List = [None]
        if not self.shared_viewer:
            self.viewers = [None] * self.n
        self._reset_render()

    def step(
        self, action_n: Dict[str, Union[int, List[float]]]
    ) -> Tuple[
        Dict[str, Union[np.array, Any]],
        Union[dict, Dict[str, Union[float, Any]]],
        Dict[str, Any],
        Dict[str, dict],
    ]:
        obs_n = {}
        reward_n = {}
        done_n = {}
        # set action for each agent
        for agent_id in self.agent_ids:
            agent = self.agents[agent_id]
            agent_action = copy.deepcopy(action_n[agent_id])
            self._set_action(agent_action, agent, self.action_spaces[agent_id])
        # advance world state
        self.world.step()
        # record observation for each agent
        for a_i, agent_id in enumerate(self.agent_ids):
            agent = self.agents[agent_id]
            obs_n[agent_id] = self._get_obs(a_i, agent)
            reward_n[agent_id] = self._get_reward(a_i, agent)
            done_n[agent_id] = self._get_done(agent)

            if done_n[agent_id]:
                self.env_done = True

        state_n = self._get_state()

        return obs_n, reward_n, done_n, state_n

    def reset(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        # reset world
        if self.reset_callback is not None:
            self.reset_callback(self.world)
        else:
            raise ValueError("self.reset_callback is still None!")
        # reset renderer
        self._reset_render()
        self.env_done = False
        # record observations for each agent
        obs_n = {}
        for a_i, agent_id in enumerate(self.agent_ids):
            agent = self.agents[agent_id]
            obs_n[agent_id] = self._get_obs(a_i, agent)
        state_n = self._get_state()
        return obs_n, {"s_t": state_n}

    # get info used for benchmarking
    def _get_info(self, agent: Agent) -> Dict:
        if self.info_callback is None:
            return {}
        return self.info_callback(agent, self.world)

    # get observation for a particular agent
    def _get_obs(self, a_i: int, agent: Agent) -> np.array:
        if self.observation_callback is None:
            return np.zeros(0)
        else:
            return self.observation_callback(agent, a_i, self.world)

    # get dones for a particular agent
    # unused right now -- agents are allowed to go beyond the viewing screen
    def _get_done(self, agent: Agent) -> bool:
        if self.done_callback is None:
            raise ValueError("No done callback specified.")
        return self.done_callback(agent, self.world)

    # get reward for a particular agent
    def _get_reward(self, a_i: int, agent: Agent) -> float:
        if self.reward_callback is None:
            return 0.0
        return self.reward_callback(agent, a_i, self.world)

    def _get_state(self) -> np.array:
        # get positions of all entities in this agent's reference frame
        entity_pos = []
        for entity in self.world.landmarks:  # world.entities:
            entity_pos.append(entity.state.p_pos)
        # entity colors

        agent_pos = []
        agent_vel = []
        for i, agent in enumerate(self.world.agents):
            agent_pos.append(agent.state.p_pos)
            agent_vel.append(agent.state.p_vel)

        return np.array(
            np.concatenate(
                [[self.world.current_step / 50]] + entity_pos + agent_pos + agent_vel
            ),
            dtype=np.float32,
        )

    # set env action for a particular agent
    def _set_action(
        self,
        action: np.array,
        agent: Agent,
        action_space: spaces,
    ) -> None:
        agent.action.u = np.zeros(self.world.dim_p)
        # process action
        if isinstance(action_space, MultiDiscrete):
            act = []
            size = action_space.high - action_space.low + 1
            index = 0
            for s in size:
                act.append(action[index : (index + s)])
                index += s
            action = act
        else:
            action = [action]

        if agent.movable:
            # physical action
            if self.discrete_action_input:
                agent.action.u = np.zeros(self.world.dim_p)
                # process discrete action
                if action[0] == 1:
                    agent.action.u[0] = -1.0
                if action[0] == 2:
                    agent.action.u[0] = +1.0
                if action[0] == 3:
                    agent.action.u[1] = -1.0
                if action[0] == 4:
                    agent.action.u[1] = +1.0
            else:
                # if self.force_discrete_action:
                #     d = np.argmax(action[0])
                #     action[0][:] = 0.0
                #     action[0][d] = 1.0
                if self.discrete_action_space:
                    agent.action.u[0] += action[0][1] - action[0][2]
                    agent.action.u[1] += action[0][3] - action[0][4]
                else:
                    agent.action.u = action[0]
            sensitivity = 5.0
            if agent.accel is not None:
                sensitivity = agent.accel
            agent.action.u *= sensitivity
            action = action[1:]
        # make sure we used all elements of action
        assert len(action) == 0

    # reset rendering assets
    def _reset_render(self) -> None:
        self.render_geoms = None
        self.render_geoms_xform = None

    # render environment
    def render(self, mode: str = "human") -> List[np.array]:

        for i in range(len(self.viewers)):
            # create viewers (if necessary)
            if self.viewers[i] is None:
                # import rendering only if we need it (and don't
                # import for headless machines)
                # from gym.envs.classic_control import rendering
                from . import rendering

                self.viewers[i] = rendering.Viewer(700, 700)

        # create rendering geometry
        if self.render_geoms is None:
            # import rendering only if we need it (and don't import
            # for headless machines)
            # from gym.envs.classic_control import rendering
            from . import rendering

            self.render_geoms = []
            self.render_geoms_xform = []
            for entity in self.world.entities:
                geom = rendering.make_circle(entity.size)
                xform = rendering.Transform()

                if entity.color is not None:
                    r, g, b = entity.color
                    if "agent" in entity.name:
                        geom.set_color(r, g, b, alpha=0.5)
                    else:
                        geom.set_color(r, g, b)
                geom.add_attr(xform)
                self.render_geoms.append(geom)
                self.render_geoms_xform.append(xform)

            # add geoms to viewer
            for viewer in self.viewers:
                viewer.geoms = []
                for geom in self.render_geoms:
                    viewer.add_geom(geom)

        results = []
        for i in range(len(self.viewers)):
            # update bounds to center around agent
            cam_range = 1
            if self.shared_viewer:
                pos = np.zeros(self.world.dim_p)
            else:
                pos = self.agents[self.agent_ids[i]].state.p_pos
            self.viewers[i].set_bounds(
                pos[0] - cam_range,
                pos[0] + cam_range,
                pos[1] - cam_range,
                pos[1] + cam_range,
            )
            # update geometry positions
            if self.render_geoms_xform is not None:
                for e, entity in enumerate(self.world.entities):
                    self.render_geoms_xform[e].set_translation(*entity.state.p_pos)
            else:
                raise ValueError("self.render_geoms_xform is still None!")
            # render to display or array
            results.append(self.viewers[i].render(return_rgb_array=mode == "rgb_array"))

        return np.squeeze(results)

    # create receptor field locations in local coordinate frame
    def _make_receptor_locations(self) -> List[np.array]:
        receptor_type = "polar"
        range_min = 0.05 * 2.0
        range_max = 1.00
        dx = []
        # circular receptive field
        if receptor_type == "polar":
            for angle in np.linspace(-np.pi, +np.pi, 8, endpoint=False):
                for distance in np.linspace(range_min, range_max, 3):
                    dx.append(distance * np.array([np.cos(angle), np.sin(angle)]))
            # add origin
            dx.append(np.array([0.0, 0.0]))
        # grid receptive field
        if receptor_type == "grid":
            for x in np.linspace(-range_max, +range_max, 5):
                for y in np.linspace(-range_max, +range_max, 5):
                    dx.append(np.array([x, y]))
        return dx



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adders for Reverb replay buffers."""

# pylint: disable=unused-import

from mava.adders.reverb.base import (
    DEFAULT_PRIORITY_TABLE,
    PriorityFn,
    PriorityFnInput,
    ReverbParallelAdder,
    Step,
)
from mava.adders.reverb.episode import ParallelEpisodeAdder
from mava.adders.reverb.sequence import ParallelSequenceAdder
from mava.adders.reverb.transition import ParallelNStepTransitionAdder



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic environment loop wrapper to track system statistics"""

import time
from typing import Dict, List, Tuple, Union

import dm_env
import matplotlib.pyplot as plt
import numpy as np
from acme.utils import counting, loggers, paths
from acme.wrappers.video import _make_animation
from array2gif import write_gif

import mava
from mava.environment_loop import ParallelEnvironmentLoop
from mava.utils.loggers import Logger
from mava.utils.wrapper_utils import RunningStatistics


class EnvironmentLoopStatisticsBase(ParallelEnvironmentLoop):
    """A parallel MARL environment loop.
    This takes `Environment` and `Executor` instances and coordinates their
    interaction. Executors are updated if `should_update=True`. This can be used as:
        loop = EnvironmentLoop(environment, executor)
        loop.run(num_episodes)
    A `Counter` instance can optionally be given in order to maintain counts
    between different Mava components. If not given a local Counter will be
    created to maintain counts between calls to the `run` method.
    A `Logger` instance can also be passed in order to control the output of the
    loop. If not given a platform-specific default logger will be used as defined
    by utils.loggers.make_default_logger from acme. A string `label` can be passed
    to easily change the label associated with the default logger; this is ignored
    if a `Logger` instance is given.
    """

    def __init__(
        self,
        environment_loop: ParallelEnvironmentLoop,
    ) -> None:
        self._environment = environment_loop._environment
        self._executor = environment_loop._executor
        self._counter = environment_loop._counter
        self._logger = environment_loop._logger
        self._should_update = environment_loop._should_update

    def _compute_step_statistics(self, rewards: Dict[str, float]) -> None:
        raise NotImplementedError

    def _compute_episode_statistics(
        self,
        episode_returns: Dict[str, float],
        episode_steps: int,
        start_time: float,
    ) -> None:
        raise NotImplementedError


class DetailedEpisodeStatistics(EnvironmentLoopStatisticsBase):
    def __init__(
        self,
        environment_loop: ParallelEnvironmentLoop,
        summary_stats: List = ["mean", "max", "min", "var", "std", "raw"],
    ):
        super().__init__(environment_loop)
        self._summary_stats = summary_stats
        self._metrics = ["episode_length", "episode_return", "steps_per_second"]
        self._running_statistics: Dict[str, float] = {}
        for metric in self._metrics:
            self.__setattr__(f"_{metric}_stats", RunningStatistics(metric))
            for stat in self._summary_stats:
                self._running_statistics[f"{stat}_{metric}"] = 0.0

    def _compute_step_statistics(self, rewards: Dict[str, float]) -> None:
        pass

    def _compute_episode_statistics(
        self,
        episode_returns: Dict[str, float],
        episode_steps: int,
        start_time: float,
    ) -> None:

        # Collect the results and combine with counts.
        steps_per_second = episode_steps / (time.time() - start_time)
        mean_episode_return = np.mean(np.array(list(episode_returns.values())))

        # Record counts.
        counts = self._counter.increment(episodes=1, steps=episode_steps)

        self._episode_length_stats.push(episode_steps)
        self._episode_return_stats.push(mean_episode_return)
        self._steps_per_second_stats.push(steps_per_second)

        for metric in self._metrics:
            for stat in self._summary_stats:
                self._running_statistics[f"{stat}_{metric}"] = self.__getattribute__(
                    f"_{metric}_stats"
                ).__getattribute__(stat)()

        self._running_statistics.update({"episode_length": episode_steps})
        self._running_statistics.update(counts)


class DetailedPerAgentStatistics(DetailedEpisodeStatistics):
    def __init__(self, environment_loop: ParallelEnvironmentLoop):
        super().__init__(environment_loop)

        # get loop logger data
        loop_label = self._logger._label
        base_dir = self._logger._directory
        (
            to_terminal,
            to_csv,
            to_tensorboard,
            time_delta,
            print_fn,
            time_stamp,
        ) = self._logger._logger_info

        self._agents_stats: Dict[str, Dict[str, RunningStatistics]] = {
            agent: {} for agent in self._environment.possible_agents
        }
        self._agent_loggers: Dict[str, loggers.Logger] = {}

        # statistics dictionary
        for agent in self._environment.possible_agents:
            agent_label = loop_label + "_" + agent
            self._agent_loggers[agent] = Logger(
                label=agent_label,
                directory=base_dir,
                to_terminal=to_terminal,
                to_csv=to_csv,
                to_tensorboard=to_tensorboard,
                time_delta=time_delta,
                print_fn=print_fn,
                time_stamp=time_stamp,
            )
            self._agents_stats[agent]["return"] = RunningStatistics(
                f"{agent}_episode_return"
            )
            self._agents_stats[agent]["reward"] = RunningStatistics(
                f"{agent}_episode_reward"
            )

    def _compute_step_statistics(self, rewards: Dict[str, float]) -> None:
        for agent, reward in rewards.items():
            agent_running_statistics: Dict[str, float] = {}
            self._agents_stats[agent]["reward"].push(reward)
            for stat in self._summary_stats:
                agent_running_statistics[
                    f"{agent}_{stat}_step_reward"
                ] = self._agents_stats[agent]["reward"].__getattribute__(stat)()
            self._agent_loggers[agent].write(agent_running_statistics)

    def _compute_episode_statistics(
        self,
        episode_returns: Dict[str, float],
        episode_steps: int,
        start_time: float,
    ) -> None:

        # Collect the results and combine with counts.
        steps_per_second = episode_steps / (time.time() - start_time)
        mean_episode_return = np.mean(np.array(list(episode_returns.values())))

        # Record counts.
        counts = self._counter.increment(episodes=1, steps=episode_steps)

        self._episode_length_stats.push(episode_steps)
        self._episode_return_stats.push(mean_episode_return)
        self._steps_per_second_stats.push(steps_per_second)

        for metric in self._metrics:
            for stat in self._summary_stats:
                self._running_statistics[f"{stat}_{metric}"] = self.__getattribute__(
                    f"_{metric}_stats"
                ).__getattribute__(stat)()

        for agent, agent_return in episode_returns.items():
            agent_running_statistics: Dict[str, float] = {}
            self._agents_stats[agent]["return"].push(agent_return)
            for stat in self._summary_stats:
                agent_running_statistics[f"{agent}_{stat}_return"] = self._agents_stats[
                    agent
                ]["return"].__getattribute__(stat)()
            self._agent_loggers[agent].write(agent_running_statistics)

        self._running_statistics.update({"episode_length": episode_steps})
        self._running_statistics.update(counts)


class MonitorParallelEnvironmentLoop(ParallelEnvironmentLoop):
    """A MARL environment loop.
    This records a gif/video every `record_every` eval episodes.
    """

    def __init__(
        self,
        environment: dm_env.Environment,
        executor: mava.core.Executor,
        filename: str = "agents",
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        should_update: bool = True,
        label: str = "parallel_environment_loop",
        record_every: int = 1000,
        path: str = "~/mava",
        fps: int = 15,
        counter_str: str = "evaluator_episodes",
        format: str = "video",
        figsize: Union[float, Tuple[int, int]] = (360, 640),
    ):
        assert (
            format == "gif" or format == "video"
        ), "Only gif and video format are supported."

        # Internalize agent and environment.
        super().__init__(
            environment=environment,
            executor=executor,
            counter=counter,
            logger=logger,
            should_update=should_update,
            label=label,
        )
        self._record_every = record_every
        self._path = paths.process_path(path, "recordings", add_uid=False)
        self._filename = filename
        self._record_current_episode = False
        self._fps = fps
        self._frames: List = []

        self._parent_environment_step = self._environment.step
        self._environment.step = self.step

        self._parent_environment_reset = self._environment.reset
        self._environment.reset = self.reset

        self._counter_str = counter_str

        self._format = format
        self._figsize = figsize

    def step(self, action: Dict[str, np.ndarray]) -> dm_env.TimeStep:
        timestep = self._parent_environment_step(action)
        self._append_frame()
        return timestep

    def _retrieve_render(self) -> np.ndarray:
        render = None
        try:
            if self._format == "video":
                render = self._environment.render(mode="rgb_array")
            elif self._format == "gif":
                render = np.transpose(
                    self._environment.render(mode="rgb_array"), axes=(1, 0, 2)
                )
        except Exception as ex:
            print(f"Render frames exception: {ex}")
            pass
        return render

    def _append_frame(self) -> None:
        """Appends a frame to the sequence of frames."""
        counts = self._counter.get_counts()
        counter = counts.get(self._counter_str)
        if counter and (counter % self._record_every == 0):
            self._frames.append(self._retrieve_render())

    def reset(self) -> dm_env.TimeStep:
        if self._frames:
            self._write_frames()
        timestep = self._parent_environment_reset()
        return timestep

    def _write_frames(self) -> None:
        counts = self._counter.get_counts()
        counter = counts.get(self._counter_str)
        path = f"{self._path}/{self._filename}_{counter}_eval_episode"
        try:
            if self._format == "video":
                self._save_video(path)
            elif self._format == "gif":
                self._save_gif(path)
        except Exception as ex:
            print(f"Write frames exception: {ex}")
            pass
        self._frames = []
        # Clear matplotlib figures in memory
        plt.close("all")

    def _save_video(self, path: str) -> None:
        video = _make_animation(self._frames, self._fps, self._figsize).to_html5_video()
        with open(f"{path}.html", "w") as f:
            f.write(video)

    def _save_gif(self, path: str) -> None:
        write_gif(
            self._frames,
            f"{path}.gif",
            fps=self._fps,
        )



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO (StJohn):
#   - [] complete base class for mixing

import abc
from typing import Dict, Optional

import sonnet as snt

from mava import specs as mava_specs
from mava.components.tf.architectures import BaseArchitecture

"""Base mixing interface for multi-agent RL systems"""


class BaseMixingModule:
    """Base class for MARL mixing.
    Objects which implement this interface provide a set of functions
    to create systems that can perform value decomposition via a mixing
    strategy between agents in a multi-agent RL system.
    """

    @abc.abstractmethod
    def __init__(
        self,
        architecture: Optional[BaseArchitecture] = None,
        environment_spec: Optional[mava_specs.MAEnvironmentSpec] = None,
        agent_networks: Optional[Dict[str, snt.Module]] = None,
    ) -> None:
        """Initialise the mixer."""

    @abc.abstractmethod
    def _create_mixing_layer(self) -> snt.Module:
        """Abstract function for adding an arbitrary mixing layer to a
        given architecture."""

    @abc.abstractmethod
    def create_system(self) -> Dict[str, Dict[str, snt.Module]]:
        """Create/update system architecture with specified mixing."""
